/***************************************************************************
 * Video Poker                                                             *
 * written for the Game Boy Advance                                        *
 * (c) 2003 Nathan Rosenquist                                              *
 *                                                                         *
 * https://nathanrosenquist.com/gba-video-poker/                           *
 **************************************************************************/

/* GBA HARDWARE INCLUDES */
#include "include/gba.h"
#include "include/screenmode.h"
#include "include/keypad.h"
#include "include/timers.h"
#include "include/qrand.h"

/* BITMAP IMAGES */
#include "images/palette.h"
#include "images/black.h"
#include "images/titlescreen.h"
#include "images/gameover.h"
#include "images/background.h"
#include "images/payout_table.h"
#include "images/payout_table_win.h"
#include "images/cards/hearts.h"
#include "images/cards/diamonds.h"
#include "images/cards/clubs.h"
#include "images/cards/spades.h"
#include "images/selecter.h"
#include "images/hold.h"
#include "images/credits.h"

/* GENERIC CONSTANTS */
#define TRUE    1
#define FALSE   0
#define NULL    0

/* GRAPHICS CONSTANTS */
#define SCREEN_WIDTH        240
#define SCREEN_HEIGHT       160
#define HALF_SCREEN_WIDTH   120

#define CARDS_Y_OFFSET       75

#define FrontBuffer     (u16*)0x6000000
#define BackBuffer      (u16*)0x600A000 /* start of back buffer in Mode 4 */

#define RGB16(r,g,b)    ((r)+(g<<5)+(b<<10))

/* GAME MODES */
#define WAIT_FOR_GAME_START 0
#define WAIT_FOR_DEAL       1
#define WAIT_FOR_REDEAL     2
#define GAME_OVER           3

/*******************/
/* POKER CONSTANTS */
/*******************/
#define BET_AMOUNT       5
#define CARDS_IN_DECK   52
#define CARDS_IN_HAND    5

/* SUITS */
#define HEARTS      0
#define DIAMONDS    1
#define CLUBS       2
#define SPADES      3

/* FACE CARDS */
#define ACE         1
#define JACK        11
#define QUEEN       12
#define KING        13

/* WINNING HANDS */
#define LOSE            0
#define JACKS_OR_BETTER 1
#define TWO_PAIR        2
#define THREE_OF_A_KIND 3
#define STRAIGHT        4
#define FLUSH           5
#define FULL_HOUSE      6
#define FOUR_OF_A_KIND  7
#define STRAIGHT_FLUSH  8
#define ROYAL_FLUSH     9

/* PAYOUT TABLE */
#define PAYOUT_ROYAL_FLUSH      4000
#define PAYOUT_STRAIGHT_FLUSH    250
#define PAYOUT_FOUR_OF_A_KIND    125
#define PAYOUT_FULL_HOUSE         45
#define PAYOUT_FLUSH              30
#define PAYOUT_STRAIGHT           20
#define PAYOUT_THREE_OF_A_KIND    15
#define PAYOUT_TWO_PAIR           10
#define PAYOUT_JACKS_OR_BETTER     5

/*********************/
/* DECLARE VARIABLES */
/*********************/

/* graphics buffer and palette pointers */
u16 *ScreenBuffer   = (u16*)BackBuffer;
u16 *ScreenPalette  = (u16*)BGPaletteMem;

/* card structure */
typedef struct {
    char suit;
    char number;
    u16 *image;                 /* pointer to bitmap from images/cards.h */
} card;

/* card arrays */
card cards[CARDS_IN_DECK];      /* one of every card, a template */
card *deck[CARDS_IN_DECK];      /* the deck, gets shuffled, etc */
card *hand[CARDS_IN_HAND];      /* player's cards */
int held[CARDS_IN_HAND];        /* TRUE/FALSE by position in hand[] array */

/* assorted global variables */
int win = LOSE;                 /* the winning condition of our hand */
int mode = WAIT_FOR_GAME_START; /* game state */
int selecterIndex = 0;          /* card the selecter is on, 0-4 */
int initialCredits = 20;        /* number of credits to start with */
int credits = 0;                /* active credit counter */
int payout = 0;                 /* the amount we can win from this hand */

/*************/
/* FUNCTIONS */
/*************/

/* create a reference structure with one of every type of card */
void createCardList(void) {
    int suit;
    int number;
    u16 *imgPtr = NULL;

    for (suit=HEARTS; suit<=SPADES; suit++) {
        for (number=ACE; number<=KING; number++) {
            cards[ ((suit*KING) + number) - 1 ].suit = suit;
            cards[ ((suit*KING) + number) - 1 ].number = number;

            if (HEARTS == suit) {
                switch (number) {
                    case ACE:
                        imgPtr = (u16*)heartsAData;
                        break;
                    case 2:
                        imgPtr = (u16*)hearts2Data;
                        break;
                    case 3:
                        imgPtr = (u16*)hearts3Data;
                        break;
                    case 4:
                        imgPtr = (u16*)hearts4Data;
                        break;
                    case 5:
                        imgPtr = (u16*)hearts5Data;
                        break;
                    case 6:
                        imgPtr = (u16*)hearts6Data;
                        break;
                    case 7:
                        imgPtr = (u16*)hearts7Data;
                        break;
                    case 8:
                        imgPtr = (u16*)hearts8Data;
                        break;
                    case 9:
                        imgPtr = (u16*)hearts9Data;
                        break;
                    case 10:
                        imgPtr = (u16*)hearts10Data;
                        break;
                    case JACK:
                        imgPtr = (u16*)heartsJData;
                        break;
                    case QUEEN:
                        imgPtr = (u16*)heartsQData;
                        break;
                    case KING:
                        imgPtr = (u16*)heartsKData;
                        break;
                }
            } else if (DIAMONDS == suit) {
                switch (number) {
                    case ACE:
                        imgPtr = (u16*)diamondsAData;
                        break;
                    case 2:
                        imgPtr = (u16*)diamonds2Data;
                        break;
                    case 3:
                        imgPtr = (u16*)diamonds3Data;
                        break;
                    case 4:
                        imgPtr = (u16*)diamonds4Data;
                        break;
                    case 5:
                        imgPtr = (u16*)diamonds5Data;
                        break;
                    case 6:
                        imgPtr = (u16*)diamonds6Data;
                        break;
                    case 7:
                        imgPtr = (u16*)diamonds7Data;
                        break;
                    case 8:
                        imgPtr = (u16*)diamonds8Data;
                        break;
                    case 9:
                        imgPtr = (u16*)diamonds9Data;
                        break;
                    case 10:
                        imgPtr = (u16*)diamonds10Data;
                        break;
                    case JACK:
                        imgPtr = (u16*)diamondsJData;
                        break;
                    case QUEEN:
                        imgPtr = (u16*)diamondsQData;
                        break;
                    case KING:
                        imgPtr = (u16*)diamondsKData;
                        break;
                }
            } else if (CLUBS == suit) {
                switch (number) {
                    case ACE:
                        imgPtr = (u16*)clubsAData;
                        break;
                    case 2:
                        imgPtr = (u16*)clubs2Data;
                        break;
                    case 3:
                        imgPtr = (u16*)clubs3Data;
                        break;
                    case 4:
                        imgPtr = (u16*)clubs4Data;
                        break;
                    case 5:
                        imgPtr = (u16*)clubs5Data;
                        break;
                    case 6:
                        imgPtr = (u16*)clubs6Data;
                        break;
                    case 7:
                        imgPtr = (u16*)clubs7Data;
                        break;
                    case 8:
                        imgPtr = (u16*)clubs8Data;
                        break;
                    case 9:
                        imgPtr = (u16*)clubs9Data;
                        break;
                    case 10:
                        imgPtr = (u16*)clubs10Data;
                        break;
                    case JACK:
                        imgPtr = (u16*)clubsJData;
                        break;
                    case QUEEN:
                        imgPtr = (u16*)clubsQData;
                        break;
                    case KING:
                        imgPtr = (u16*)clubsKData;
                        break;
                }
            } else if (SPADES == suit) {
                switch (number) {
                    case ACE:
                        imgPtr = (u16*)spadesAData;
                        break;
                    case 2:
                        imgPtr = (u16*)spades2Data;
                        break;
                    case 3:
                        imgPtr = (u16*)spades3Data;
                        break;
                    case 4:
                        imgPtr = (u16*)spades4Data;
                        break;
                    case 5:
                        imgPtr = (u16*)spades5Data;
                        break;
                    case 6:
                        imgPtr = (u16*)spades6Data;
                        break;
                    case 7:
                        imgPtr = (u16*)spades7Data;
                        break;
                    case 8:
                        imgPtr = (u16*)spades8Data;
                        break;
                    case 9:
                        imgPtr = (u16*)spades9Data;
                        break;
                    case 10:
                        imgPtr = (u16*)spades10Data;
                        break;
                    case JACK:
                        imgPtr = (u16*)spadesJData;
                        break;
                    case QUEEN:
                        imgPtr = (u16*)spadesQData;
                        break;
                    case KING:
                        imgPtr = (u16*)spadesKData;
                        break;
                }

            }

            /* assign the pointer to the card image */
            cards[ ((suit*KING) + number) - 1 ].image = imgPtr;
        }
    }
}

/* create our working deck, which has pointers to the reference cards */
void initDeck(void) {
    int i;

    for (i=0; i<CARDS_IN_DECK; i++) {
        deck[i] = &cards[i];
    }
}

/* move the pointers in the deck around randomly */
void shuffleDeck(void) {
    int i;
    int rnd = 0;
    card *tmpCard = NULL;

    /* loop through each card, pick another random card (after the first card), and swap the two */
    for (i=0; i<CARDS_IN_DECK; i++) {
        rnd = (int)(((float)CARDS_IN_DECK - (float)i) * rand()/RAND_MAX+1.0) + i - 1;

        tmpCard     = deck[i];
        deck[i]     = deck[rnd];
        deck[rnd]   = tmpCard;
    }
}

/* clear all hold flags on cards */
void resetHeldCards(void) {
    int i;

    for (i=0; i<CARDS_IN_HAND; i++) {
        held[i] = FALSE;
    }
}

/* take cards from the deck and copy them into the user's hand */
/* if it's a re-deal, start at card 6 */
/* this would need to be reworked if we had multiple players */
void deal(int *heldCards) {
    int i = 0;
    int nextIndex = CARDS_IN_HAND;

    /* original deal */
    if (NULL == heldCards) {
        for (i=0; i<CARDS_IN_HAND; i++) {
            hand[i] = deck[i];
        }
    /* re-deal */
    } else {
        for (i=0; i<CARDS_IN_HAND; i++) {
            if (FALSE == held[i]) {
                hand[i] = deck[nextIndex];
                nextIndex++;
            }
        }
    }

    resetHeldCards();
}

/****************************/
/* POKER SPECIFIC FUNCTIONS */
/****************************/

int isFaceCard(int number) {
    if ((number > 10) || (ACE == number)) {
        return TRUE;
    }

    return FALSE;
}

int isJacksOrBetter(void) {
    int i;
    int j;

    for (i=0; i<CARDS_IN_HAND; i++) {
        if (isFaceCard(hand[i]->number)) {
            for (j=0; j<CARDS_IN_HAND; j++) {
                if (i != j) {
                    if (hand[i]->number == hand[j]->number) {
                        if (isFaceCard(hand[i]->number) && isFaceCard(hand[i]->number)) {
                            return TRUE;
                        }
                    }
                }
            }
        }
    }

    return FALSE;
}

int isTwoPair(void) {
    int pair1Index1 = -1;
    int pair1Index2 = -1;
    int i;
    int j;

    /* pair #1 */
    for (i=0; i<CARDS_IN_HAND; i++) {
        for (j=0; j<CARDS_IN_HAND; j++) {
            if (i != j) {
                if (hand[i]->number == hand[j]->number) {
                    pair1Index1 = i;
                    pair1Index2 = j;
                }
            }
        }
    }

    /* if we have one pair, find pair #2 */
    if (-1 != pair1Index1) {
        for (i=0; i<CARDS_IN_HAND; i++) {
            for (j=0; j<CARDS_IN_HAND; j++) {
                if (i != j) {
                    if ((i != pair1Index1) && (i != pair1Index2) && (j != pair1Index1) && (j != pair1Index2)) {
                        if (hand[i]->number == hand[j]->number) {
                            return TRUE;
                        }
                    }
                }
            }
        }
    }

    return FALSE;
}

int isThreeOfAKind(void) {
    int i;
    int j;
    int k;

    for (i=0; i<CARDS_IN_HAND; i++) {
        for (j=0; j<CARDS_IN_HAND; j++) {
            if (i != j) {
                for (k=0; k<CARDS_IN_HAND; k++) {
                    if ((i != k) && (j != k)) {
                        if (hand[i]->number == hand[j]->number) {
                            if (hand[j]->number == hand[k]->number) {
                                return TRUE;
                            }
                        }
                    }
                }
            }
        }
    }

    return FALSE;
}

int isStraight(void) {
    int numbers[CARDS_IN_HAND];
    int tmp = 0;
    int i = 0;
    int j = 0;

    /* stick numbers in array */
    for (i=0; i<CARDS_IN_HAND; i++) {
        numbers[i] = hand[i]->number;
    }

    /* bubble sort */
    for (i=0; i<4; i++) {
        for (j=0; j<4; j++) {
            if (numbers[j] > numbers[j+1]) {
                tmp = numbers[j];
                numbers[j] = numbers[j+1];
                numbers[j+1] = tmp;
            }
        }
    }

    /* check to see if numbers are contiguous */
    if (numbers[0] == (numbers[1] - 1)) {
        if (numbers[1] == (numbers[2] - 1)) {
            if (numbers[2] == (numbers[3] - 1)) {
                if (numbers[3] == (numbers[4] - 1)) {
                    return TRUE;
                }
            }
        }
    }

    /* the ace is an exception */
    if (ACE == numbers[0]) {
        if (10 == numbers[1]) {
            if (JACK == numbers[2]) {
                if (QUEEN == numbers[3]) {
                    if (KING == numbers[4]) {
                        return TRUE;
                    }
                }
            }
        }
    }

    return FALSE;
}

int isFlush(void) {
    if (hand[0]->suit == hand[1]->suit) {
        if (hand[1]->suit == hand[2]->suit) {
            if (hand[2]->suit == hand[3]->suit) {
                if (hand[3]->suit == hand[4]->suit) {
                    return TRUE;
                }
            }
        }
    }

    return FALSE;
}

int isFullHouse(void) {
    int pairIndex1 = -1;
    int pairIndex2 = -1;
    int i;
    int j;
    int k;

    /* find the pair first */
    for (i=0; i<CARDS_IN_HAND; i++) {
        for (j=0; j<CARDS_IN_HAND; j++) {
            if (i != j) {
                if (hand[i]->number == hand[j]->number) {
                    pairIndex1 = i;
                    pairIndex2 = j;
                }
            }
        }
    }

    /* if we have a pair, check for the other three of a kind */
    if (-1 != pairIndex1) {
        for (i=0; i<CARDS_IN_HAND; i++) {
            if ((i != pairIndex1) && (i != pairIndex2)) {
                for (j=0; j<CARDS_IN_HAND; j++) {
                    if (i != j) {
                        if ((j != pairIndex1) && (j != pairIndex2)) {
                            for (k=0; k<CARDS_IN_HAND; k++) {
                                if ((i != k) && (j != k)) {
                                    if ((k != pairIndex1) && (k != pairIndex2)) {
                                        if (hand[i]->number == hand[j]->number) {
                                            if (hand[j]->number == hand[k]->number) {
                                                return TRUE;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    return FALSE;
}

int isFourOfAKind(void) {
    int i;
    int j;
    int k;
    int l;

    for (i=0; i<CARDS_IN_HAND; i++) {
        for (j=0; j<CARDS_IN_HAND; j++) {
            if (i != j) {
                for (k=0; k<CARDS_IN_HAND; k++) {
                    if ((i != k) && (j != k)) {
                        for (l=0; l<CARDS_IN_HAND; l++) {
                            if ((i != l) && (j != l) && (k != l)) {
                                if (hand[i]->number == hand[j]->number) {
                                    if (hand[j]->number == hand[k]->number) {
                                        if (hand[k]->number == hand[l]->number) {
                                            return TRUE;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    return FALSE;
}

int isStraightFlush(void) {
    if (isStraight() && isFlush()) {
        return TRUE;
    }

    return FALSE;
}

int isRoyalFlush(void) {
    if (isStraightFlush()) {
        if ((10 + JACK + QUEEN + KING + ACE) == (hand[0]->number + hand[1]->number + hand[2]->number + hand[3]->number + hand[4]->number)) {
            return TRUE;
        }
    }

    return FALSE;
}

/* sets "win" and "payout" global variables */
/* win is the type of hand, payout is the money awarded */
void checkWin(void) {
    if (isRoyalFlush()) {
        win     = ROYAL_FLUSH;
        payout  = PAYOUT_ROYAL_FLUSH;

    } else if (isStraightFlush()) {
        win     = STRAIGHT_FLUSH;
        payout  = PAYOUT_STRAIGHT_FLUSH;

    } else if (isFourOfAKind()) {
        win     = FOUR_OF_A_KIND;
        payout  = PAYOUT_FOUR_OF_A_KIND;

    } else if (isFullHouse()) {
        win     = FULL_HOUSE;
        payout  = PAYOUT_FULL_HOUSE;

    } else if (isFlush()) {
        win     = FLUSH;
        payout  = PAYOUT_FLUSH;

    } else if (isStraight()) {
        win     = STRAIGHT;
        payout  = PAYOUT_STRAIGHT;

    } else if (isThreeOfAKind()) {
        win     = THREE_OF_A_KIND;
        payout  = PAYOUT_THREE_OF_A_KIND;

    } else if (isTwoPair()) {
        win     = TWO_PAIR;
        payout  = PAYOUT_TWO_PAIR;

    } else if (isJacksOrBetter()) {
        win     = JACKS_OR_BETTER;
        payout  = PAYOUT_JACKS_OR_BETTER;

    } else {
        win     = LOSE;
        payout  = 0;
    }
}

/* the GBA doesn't have a reset button per-se, but this gets run once */
/* at startup to initialize various things */
void reset(void) {
    /* initialize card memory structures */
    createCardList();
    initDeck();
    resetHeldCards();

    /* setup graphics stuff */
    SetMode (MODE_4 | BG2_ENABLE);

    /* start the timer for entropy */
    REG_TM3CNT = TIME_ENABLE;

    /* in case we're on an emulator without timer support */
    srand(1);
}

/* alter the random number sequence according to the system timer */
void seedRandom(void) {
    if (0 != REG_TM3D) {
        srand(REG_TM3D);
    }
}

/* the GBA also has timers, but this is dead simple */
void wait(int seconds) {
    int i = 0;
    int j = 0;

    for (i=0; i<seconds; i++) {
        for (j=0; j<120000; j++) {
            /* busy-wait */
        }
    }
}

/* swap graphics buffers */
void flipBuffer(void) {
    /* wait for the right time */
    vsync();

    /* back buffer is the current buffer so we need to switch it to the front buffer */
    if(REG_DISPCNT & BACKBUFFER) {
        /* flip active buffer to front buffer by clearing back buffer bit */
        REG_DISPCNT &= ~BACKBUFFER;
        /* now we point our drawing buffer to the back buffer */
        ScreenBuffer = BackBuffer;
    /* front buffer is active so switch it to backbuffer */
    } else {
        /* flip active buffer to back buffer by setting back buffer bit */
        REG_DISPCNT |= BACKBUFFER;
        /* now we point our drawing buffer to the front buffer */
        ScreenBuffer = FrontBuffer;
    }
}

/* common function that takes full-screen image data pointer as input */
void drawFullScreen(u16 *screenData) {
    u16 x,y;
    u16 *bgPtr = screenData;
    int buffer = 0;

    for (buffer=0; buffer<2; buffer++) {
        for (x=0; x<HALF_SCREEN_WIDTH; x++) {
            for (y=0; y<SCREEN_HEIGHT; y++) {
                ScreenBuffer[x+(y*HALF_SCREEN_WIDTH) ] = bgPtr[x+(y*HALF_SCREEN_WIDTH)];
            }
        }
        flipBuffer();
    }
}

/* draws title screen */
void drawTitleScreen(void) {
    /* initialize the user's credits */
    credits = initialCredits;

    /* show the title screen */
    drawFullScreen((u16*)titleScreenData);

    /* set the mode */
    mode = WAIT_FOR_GAME_START;
}

/* game over sequence */
void gameOver(void) {
    /* wait on the previous screen for a moment */
    wait(3);

    /* show the game over screen */
    drawFullScreen((u16*)gameOverData);

    /* wait a little longer */
    wait(5);
    
    /* go back to the opening title screen */
    drawTitleScreen();
}

/* draws background photo */
void drawBackground(void) {
    u16 x,y;
    u16 *bgPtr;

    bgPtr = (u16*)backgroundData;

    for (x=0; x<HALF_SCREEN_WIDTH; x++) {
        for (y=0; y<SCREEN_HEIGHT; y++) {
            ScreenBuffer[x+(y*HALF_SCREEN_WIDTH) ] = bgPtr[x+(y*HALF_SCREEN_WIDTH)];
        }
    }
}

/* converts an integer to ASCII */
char *itoa(int i) {
    /* 10 digits + NULL */
    static char buf[11];
    char *pos = buf + sizeof(buf) - 1;
    unsigned int u;

    u = i;

    /* set the trailing null first */
    *pos = NULL;

    /* now start building the string, from right to left */
    do {
        *--pos = '0' + (u % 10);
        u /= 10;
    } while (u);

    return pos;
}

/* overwrites the data in the credits area with the payout_table bitmap */
void clearCredits(void) {
    /* this uses the same magic numbers commented in drawCreditNumber() below */
    /* jeez this is ugly! :) */

    u16 x, y;
    u16 x_offset, y_offset;
    u16 *typecast16;

    typecast16 = (u16*)payout_tableData;

    x_offset = 83; /* 83 = (166 / 2) */
    y_offset = 56;

    for (x=0; x<33; x++) { /* 33 = (66/2) */
        for (y=0; y<10; y++) {
            /* 81 = (x_offset - 2 (4px padding) ... 52 = (y_offset - 4px padding) */
            ScreenBuffer[(x + x_offset) + ((y + y_offset) * HALF_SCREEN_WIDTH)] =
                typecast16[(x + 81) + ((y + 52) * 116)];
        }
    }
}

/* draw an individual number (or $) in the credits field */
void drawCreditNumber(char num, int pos) {
    /* offsets: 166px X - 56px Y */
    /* up to 66px wide, always 10px high */
    /* right aligned, ending at 222px X - 76px Y */
    /* each number is 6px wide with no padding in between */
    /* the bitmap itself has the numbers in this order: 0123456789$ */

    u16 x, y;
    u16 x_scr_offset;
    u16 x_bmp_offset;
    u16 *typecast16;

    typecast16 = (u16*)creditsData;

    if ('$' == num) {
        x_bmp_offset = 30;  /* 30 = (60/2) */
    } else {
        /* subtracting 48 from an ascii number 0-9 gives us that number as an int */
        /* 3 = (6px width / 2) */
        x_bmp_offset = ((num - 48) * 3);
    }

    x_scr_offset = 113 - (pos * 3);

    for (x=0; x<3; x++) { /* 3 = (6/2) */
        for (y=0; y<10; y++) {
            ScreenBuffer[(x + x_scr_offset) + ((y + 56) * HALF_SCREEN_WIDTH)] =
                typecast16[(x + x_bmp_offset) + (y * 33)];
        }
    }
}

/* higher level function, draws the number of credits to the screen */
void drawCredits(void) {
    char *string = NULL;    /* string representation of the number of credits */
    int numPos = 0;         /* the numbers column (ones, tens, etc) */
    int strPos = 0;         /* string index (counting backwards to zero) */

    clearCredits();

    string = itoa(credits);

    while (string[strPos] != NULL) {
        strPos++;
    }

    /* we don't want the null bit */
    strPos--;

    /* loop through the string backwards */
    while (strPos >= 0) {
        drawCreditNumber(string[strPos], numPos);
        numPos++;
        strPos--;
    }

    /* put the dollar sign on when we're done */
    drawCreditNumber('$', numPos);
}

/* handle the bookkeeping, then call drawCredits() */
void drawWinCredits(void) {
    int buffer = 0;

    /* add the winnings to our credits */
    credits += payout;

    /* we only win what we won once */
    payout = 0;

    /* display credits on screen */
    for (buffer=0; buffer<2; buffer++) {
        drawCredits();
    }
}

/* draws the payout table at the top of the screen */
/* accepts the "win" variable as an argument */
/* if we won anything, highlight that hand in yellow */
void drawPayoutTable(int theWin) {
    /*
     * the entire payout table is 232px wide, 65px tall
     * it starts 4px down and right from the top corner of the screen
     */

    u16 x, y;
    u16 x_offset, y_offset;
    u16 *typecast16;
    int highlightBarHalfWidth = 0;

    typecast16 = (u16*)payout_tableData;

    x_offset = 2; /* 2 = (4/2) */
    y_offset = 4;

    /* 116 = (232 / 2) */
    for (x=0; x<116; x++) {
        for (y=0; y<65; y++) {
            ScreenBuffer[(x + x_offset) + ((y + y_offset) * HALF_SCREEN_WIDTH)] = typecast16[x + (y * 116)];
        }
    }

    /* we just overwrote the credits, so redraw them too */
    drawCredits();

    /* if we've won something, highlight it in the payout table */
    if (theWin != LOSE) {
        /* the highlight bar on the left size is 110px wide (51)
         * the highlight bar on the right side is 120px wide (56)
         * they are both 10px tall
         * below are the offsets for each hand */

        /***************************************/
        /* HAND               X     X/2      Y */
        /***************************************/
        /* ROYAL_FLUSH        4      2       4 */
        /* STRAIGHT_FLUSH     4      2      16 */
        /* FOUR_OF_A_KIND     4      2      28 */
        /* FULL_HOUSE         4      2      40 */
        /* FLUSH              4      2      52 */
        /* STRAIGHT         116     58       4 */
        /* THREE_OF_A_KIND  116     58      16 */
        /* TWO_PAIR         116     58      28 */
        /* JACKS_OR_BETTER  116     58      40 */
        /***************************************/

        /* figure out which offsets apply to this hand */
        if (ROYAL_FLUSH == theWin) {
            x_offset =  2;
            y_offset =  4;
            highlightBarHalfWidth = 51;
        } else if (STRAIGHT_FLUSH == theWin) {
            x_offset =  2;
            y_offset = 16;
            highlightBarHalfWidth = 51;
        } else if (FOUR_OF_A_KIND == theWin) {
            x_offset =  2;
            y_offset = 28;
            highlightBarHalfWidth = 51;
        } else if (FULL_HOUSE == theWin) {
            x_offset =  2;
            y_offset = 40;
            highlightBarHalfWidth = 51;
        } else if (FLUSH == theWin) {
            x_offset =  2;
            y_offset = 52;
            highlightBarHalfWidth = 51;
        } else if (STRAIGHT == theWin) {
            x_offset = 58;
            y_offset =  4;
            highlightBarHalfWidth = 56;
        } else if (THREE_OF_A_KIND == theWin) {
            x_offset = 58;
            y_offset = 16;
            highlightBarHalfWidth = 56;
        } else if (TWO_PAIR == theWin) {
            x_offset = 58;
            y_offset = 28;
            highlightBarHalfWidth = 56;
        } else if (JACKS_OR_BETTER == theWin) {
            x_offset = 58;
            y_offset = 40;
            highlightBarHalfWidth = 56;
        }

        /* get the pointer to the highlight image */
        typecast16 = (u16*)payout_table_winData;

        /* draw the highlight */
        for (x=0; x < highlightBarHalfWidth; x++) {
            for (y=0; y<10; y++) {
                /* the extra x2 and y4 are the payout table's offset from the main screen */
                ScreenBuffer[(x + x_offset + 2) + ((y + y_offset + 4) * HALF_SCREEN_WIDTH)] = typecast16[(x + x_offset) + ((y + y_offset) * 116)];
            }
        }
    }
}

/* draw one card to the screen */
/* cardNum specifies which card from the players hand (0-4) */
void drawCard(int cardNum) {
    /*
     * we are using mode 4, so all x values are cut in half
     * since there are no hardware divides, we use magic numbers and heavy comments :)
     *
     * poker cards are 40px wide and 60px tall
     * 4px on the left and right of the screen for padding
     * 8px between each card to space them out
     * vertically cards start at 75px down the screen
     */

    u16 x,y;
    u16 x_offset, y_offset;
    u16 *cardPtr;

    cardPtr = (u16*)hand[cardNum]->image;

    /* ((4px padding + (cardNum * (40px card + 8px spacing))) / 2) */
    x_offset = 2 + (cardNum * (20+4));
    y_offset = CARDS_Y_OFFSET;

    /* 20 == (40px card / 2) */
    for (x=0; x<20; x++) {
        for (y=0; y<60; y++) {
            ScreenBuffer[(x + x_offset) + ((y + y_offset) * HALF_SCREEN_WIDTH) ] = cardPtr[x+(y*20)];
        }
    }
}

/* draw the appropriate background bitmap section over the given card */
/* cardNum specifies the card in the player's hand (0-4) */
void clearCard(int cardNum) {
    u16 x, y;
    u16 x_offset, y_offset;
    u16 *typecast16;

    typecast16 = (u16*)backgroundData;

    x_offset = 1 + (cardNum * 24);
    y_offset = CARDS_Y_OFFSET - 2;

    for (x=0; x<22; x++) {
        for (y=0; y<64; y++) {
            ScreenBuffer[(x+x_offset) + ((y+y_offset) * HALF_SCREEN_WIDTH) ] =
                typecast16[(x+x_offset) + ((y+y_offset) * HALF_SCREEN_WIDTH)];
        }
    }
}

/* draw the thing that allows the player to select cards */
/* it's a yellow border around the cards, controlled by the joystick */
/* since this only shows up when the player's cards are on the screen, */
/* it re-draws the appropriate card over it after the selcter is drawn */
/* cardNum is the card in the player's hand (0-4) */
void drawSelecter(int cardNum) {
    /* the selecter is 44px wide and 64px tall */

    u16 x,y;
    u16 x_offset, y_offset;
    u16 *typecast16;

    typecast16 = (u16*)selecterData;

    x_offset = 1 + (cardNum * 24);
    y_offset = CARDS_Y_OFFSET - 2;

    /* 22 = (44px wide / 2) */
    for (x=0; x<22; x++) {
        for (y=0; y<64; y++) {
            ScreenBuffer[(x + x_offset) + ((y + y_offset) * HALF_SCREEN_WIDTH) ] = typecast16[x+(y*22)];
        }
    }

    drawCard(cardNum);
}

/* the selecter can move left and right when the player is selecting cards */
/* "direction" specifies the direction the selecter should move */
/* 1 is right, and -1 is left */
/* if the selecter can't move any farther left or right, it just stays put */
void moveSelecter(int direction) {
    int buffer      = 0;
    int cardNum     = 0;
    int prevCardNum = 0;

    /* moving left */
    if (-1 == direction) {
        if (selecterIndex > 0) {
            prevCardNum = selecterIndex;
            selecterIndex--;
            cardNum = selecterIndex;
        } else {
            cardNum = 0;
            prevCardNum = 1;
        }

    /* moving right */
    } else if (1 == direction) {
        if (selecterIndex < 4) {
            prevCardNum = selecterIndex;
            selecterIndex++;
            cardNum = selecterIndex;
        } else {
            cardNum = 4;
            prevCardNum = 3;
        }
    }

    /* draw twice for double buffers */
    for (buffer=0; buffer<2; buffer++) {
        drawSelecter(cardNum);
        drawCard(cardNum);
        clearCard(prevCardNum);
        drawCard(prevCardNum);
        flipBuffer();
    }
}

/* deal cards at the start of a game (not a re-deal) */
/* handles business logic and graphic presentation */
void drawDeal(void) {
    int i = 0;
    int buffer = 0;

    /* we can't be a winner yet */
    win = LOSE;

    credits -= BET_AMOUNT;

    shuffleDeck();

    /************/
    /* CHEATING */
    /************/
    /*
    deck[0] = &cards[ACE - 1];
    deck[1] = &cards[10 - 1];
    deck[2] = &cards[JACK - 1];
    deck[3] = &cards[QUEEN - 1];
    deck[4] = &cards[KING - 1];
    */

    /* pull the cards from the deck into the player's hand */
    deal(NULL);

    /* see if we were dealt a winning hand */
    checkWin();

    /* draw background */
    for (buffer=0; buffer<2; buffer++) {
        drawBackground();
        drawPayoutTable(NULL);
        flipBuffer();
    }

    /* draw cards, one at a time */
    for (buffer=0; buffer<2; buffer++) {
        for (i=0; i<CARDS_IN_HAND; i++) {
            drawCard(i);
        }
        flipBuffer();
    }

    /* now that we've seen the cards, show us if we won */
    if (win != LOSE) {
        for (buffer=0; buffer<2; buffer++) {
            drawPayoutTable(win);
            flipBuffer();
        }
    }

    /* draw selecter, and re-draw 1st card */
    selecterIndex = 0;
    for (buffer=0; buffer<2; buffer++) {
        drawSelecter(0);
        flipBuffer();
    }

    mode = WAIT_FOR_REDEAL;
}

/* does the second deal of cards (after initial cards are held or discarded) */
/* as with drawDeal(), this performs business logic and graphic presentation */
/* also seeds the RNG, as this is the best time to do it */
void drawRedeal(void) {
    int i = 0;
    int buffer = 0;

    deal(held);

    for (buffer=0; buffer<2; buffer++) {
        for (i=0; i<CARDS_IN_HAND; i++) {
            clearCard(i);
            if (FALSE == held[i]) {
                drawCard(i);
            }
        }
        flipBuffer();
    }

    /* draw the newly dealt cards */
    for (buffer=0; buffer<2; buffer++) {
        for (i=0; i<CARDS_IN_HAND; i++) {
            if (FALSE == held[i]) {
                drawCard(i);
                flipBuffer();
            }
        }
    }

    /* did we win? */
    checkWin();

    /* now show credits incrementing (if we won) */
    drawWinCredits();

    /* show win highlighted in payout table */
    for (buffer=0; buffer<2; buffer++) {
        drawPayoutTable(win);
        flipBuffer();
    }

    /* seed the RNG */
    seedRandom();

    /* if we're out of money, the game is over */
    if (0 == credits) {
        gameOver();
    }
}

/* toggles which cards are held or discarded */
/* handles business logic and graphic presentation */
/* cardNum is the card in the player's hand (0-4) */
void toggleHold(int cardNum) {
    /*
     * the hold indicator is 32px wide by 12px tall
     * it starts 4px to the right of each card, and at 142px down
     */

    int buffer = 0;
    u16 x, y;
    u16 x_offset, y_offset;
    u16 *typecast16;

    x_offset = 4 + (cardNum * 24);  /* 8 + (cardNum * 48) */
    y_offset = 142;

    /* put a hold on the card */
    if (FALSE == held[cardNum]) {
        held[cardNum] = TRUE;
        typecast16 = (u16*)holdData;

        /* 16(x) == (32 / 2) */
        for (buffer=0; buffer<2; buffer++) {
            for (x=0; x<16; x++) {
                for (y=0; y<12; y++) {
                    ScreenBuffer[(x + x_offset) + ((y + y_offset) * HALF_SCREEN_WIDTH) ] = typecast16[x+(y*16)];
                }
            }
            flipBuffer();
        }

    /* release the hold on the card */
    } else {
        held[cardNum] = FALSE;
        typecast16 = (u16*)backgroundData;

        /* 16(x) == (32 / 2) */
        for (buffer=0; buffer<2; buffer++) {
            for (x=0; x<16; x++) {
                for (y=0; y<12; y++) {
                    ScreenBuffer[(x + x_offset) + ((y + y_offset) * HALF_SCREEN_WIDTH) ] = typecast16[(x + x_offset)+((y + y_offset) * HALF_SCREEN_WIDTH)];
                }
            }
            flipBuffer();
        }
    }

}

/* loads a palette into memory */
void loadPalette(u16 *thePalette) {
    int i = 0;
    u16 *p = NULL;

    p = thePalette;

    for (i=0; i<256; i++) {
        ScreenPalette[i] = *p;
        p++;
    }
}

/* this starts the first game after the GBA is powered up */
/* this is necessary simply because this is the only time */
/* we seed the RNG before the game. also this will probably */
/* show a title screen first, etc. later */
void gameStart(void) {
    /* seed the random number generator */
    seedRandom();

    /* start the game */
    drawDeal();
}

/* main entry point into the application instead of main() */
int AgbMain(void) {
    int buffer = 0;
    int BButtonHeldDown = FALSE;

    /* run once at startup */
    reset();

    /* load the shared palette */
    loadPalette( (u16*)sharedPalette );

    /* load title screen */
    for (buffer=0; buffer<2; buffer++) {
        drawTitleScreen();
        flipBuffer();
    }

    mode = WAIT_FOR_GAME_START;

    /******************/
    /* MAIN GAME LOOP */
    /******************/
    while (TRUE) {
        /* first hand of the game, seed RNG and deal cards */
        if (WAIT_FOR_GAME_START == mode) {
            if (IS_DOWN(KEY_A) || IS_DOWN(KEY_START)) {
                gameStart();
            }
        /* wait for the user to hit deal */
        } else if (WAIT_FOR_DEAL == mode) {
            if (IS_DOWN(KEY_A)) {
                drawDeal();
            }

        /* wait for the user to hold their cards and re-deal */
        } else if (WAIT_FOR_REDEAL == mode) {
            /* move left */
            if (IS_DOWN(KEY_LEFT)) {
                moveSelecter(-1);

            /* move right */
            } else if (IS_DOWN(KEY_RIGHT)) {
                moveSelecter(1);

            /* hold card */
            } else if (IS_DOWN(KEY_B)) {
                /* showToggleHeld(selecterIndex); */
                if (FALSE == BButtonHeldDown) {
                    toggleHold(selecterIndex);
                }
                BButtonHeldDown = TRUE;
            /* re-deal */
            } else if (IS_DOWN(KEY_A)) {
                drawRedeal();
                mode = WAIT_FOR_DEAL;
            }

            if (! IS_DOWN(KEY_B)) {
                BButtonHeldDown = FALSE;
            }
        } else if (GAME_OVER == mode) {
            gameOver();
        }
    }

    /* this can't happen, but AGBMain() returns an int */
    return 0;
}

